from __future__ import absolute_import

import logging

from . import BackendBase
from ..config import BASE_MODULE_SCHEMA
from ..exceptions import KeyNotFoundException


LOG = logging.getLogger(__name__)

REQUIREMENTS = ("python-etcd",)

CONFIG_SCHEMA = BASE_MODULE_SCHEMA.copy()
CONFIG_SCHEMA.update(
    {
        "host": dict(type="string", required=True, empty=False),
        "port": dict(type="integer", default=2379, min=1, max=65535),
        "protocol": dict(type="string", default="https", empty=False),
    }
)


class Backend(BackendBase):
    def connect(self):
        import etcd

        self.etcd = etcd
        self.db = etcd.Client(
            host=self.config["host"], port=self.config["port"], protocol=self.config["protocol"]
        )

    def disconnect(self):
        LOG.debug("Disconnecting from etcd server")

    def get_val(self, key):
        LOG.debug(
            "Getting value of key %r from etcd server at %s:%s",
            key,
            self.config["host"],
            self.config["port"],
        )
        try:
            return self.db.get(key).value
        except self.etcd.EtcdKeyNotFound:
            raise KeyNotFoundException("Key %r was not found in etcd." % key)



from __future__ import absolute_import

import json
import logging

from ..config import BASE_MODULE_SCHEMA
from ..exceptions import KeyNotFoundException
from . import BackendBase

LOG = logging.getLogger(__name__)

REQUIREMENTS = ("boto3",)

CONFIG_SCHEMA = BASE_MODULE_SCHEMA.copy()
CONFIG_SCHEMA.update(
    {
        "bucket": dict(type="string", required=True, empty=False),
        "key": dict(type="string", required=True, empty=False),
    }
)

"""
{
    "version": 3,
    "terraform_version": "0.11.12",
    "serial": 65,
    "lineage": "89362db3-ded7-831f-fadd-63a32796ad81",
    "modules": [
        {
            "path": [
                "root"
            ],
            "outputs": {
                "app_access_key_id": {
                    "sensitive": false,
                    "type": "string",
                    "value": "AKIAIMUOVF5GYGI6Z3VQ"
                },
"""


class Backend(BackendBase):
    def connect(self):
        import boto3

        s3 = boto3.client("s3")
        LOG.debug(
            "Getting terraform state from s3://%s/%s",
            self.config["bucket"],
            self.config["key"],
        )
        resp = s3.get_object(Bucket=self.config["bucket"], Key=self.config["key"])
        self.state_raw = json.load(resp["Body"])
        self.state_raw["modules_dict"] = {}
        self.state = {}
        # Flatten the state file so we just have keys and values. Any modules will be
        # nested under a '<module_name>.<var_name>' key.
        for module in self.state_raw["modules"]:
            self.state_raw["modules_dict"][".".join(module["path"])] = module["outputs"]
            key = module["path"][1:]
            for name, value in module["outputs"].items():
                self.state[".".join(key + [name])] = value["value"]

    def disconnect(self):
        pass

    def get_all(self):
        return self.state_raw

    def get_val(self, key):
        try:
            return self.state[key]
        except KeyError:
            raise KeyNotFoundException("Key %r was not found in Terraform state." % key)



from __future__ import absolute_import

import logging

from . import BackendBase
from ..config import BASE_MODULE_SCHEMA
from ..exceptions import KeyNotFoundException


LOG = logging.getLogger(__name__)

REQUIREMENTS = ("redis",)

CONFIG_SCHEMA = BASE_MODULE_SCHEMA.copy()
CONFIG_SCHEMA.update(
    {
        "host": dict(type="string", required=True, empty=False),
        "port": dict(type="integer", default=6379, min=1, max=65535),
        "db": dict(type="integer", default=0, min=0, max=15),
        "password": dict(type="string", empty=False),
        "decode_responses": dict(type="boolean", default=True),
    }
)


class Backend(BackendBase):
    def connect(self):
        from redis import StrictRedis

        self.db = StrictRedis(
            host=self.config["host"],
            port=self.config["port"],
            db=self.config["db"],
            password=self.config.get("password"),
            decode_responses=self.config["decode_responses"],
        )

    def disconnect(self):
        LOG.debug("Disconnecting from Redis server")

    def get_val(self, key):
        LOG.debug(
            "Getting value of key %r from redis server at %s:%s",
            key,
            self.config["host"],
            self.config["port"],
        )
        var = self.db.get(key)
        if var is None:
            raise KeyNotFoundException("Key %r was not found in Redis." % key)
        return var



import yaml
from cerberus import Validator

from .exceptions import ConfigValidationException


BASE_MODULE_SCHEMA = {"type": dict(type="string", required=True, empty=False)}

CONFIG_SCHEMA = {
    "backends": dict(
        type="dict",
        required=True,
        keyschema=dict(type="string", regex="[a-zA-z][a-zA-Z0-9_]*"),
        valueschema=dict(type="dict", allow_unknown=True, schema=BASE_MODULE_SCHEMA),
    ),
    "templates": dict(
        type="list",
        schema={
            "src": dict(type="string", required=True, empty=False),
            "dest": dict(type="string", required=True, empty=False),
            "owner": dict(type="string", required=True, empty=False),
            "mode": dict(type="string", empty=False, default="0644"),
            "check_cmd": dict(type="string", empty=False),
            "restart_cmd": dict(type="string", empty=False),
            "vars": dict(
                type="dict",
                keyschema=dict(type="string", regex="[a-zA-z][a-zA-Z0-9_]*"),
                valueschema=dict(
                    type="dict",
                    schema={
                        "backend": dict(type="string", required=True, empty=False),
                        "key": dict(type="string", required=True, empty=False),
                        "default": dict(type="string", required=False),
                    },
                ),
            ),
        },
    ),
    "logging": dict(type="dict", allow_unknown=True, schema={}),
}


def load_config(path):
    validator = Validator(CONFIG_SCHEMA)
    with open(path) as f:
        config = yaml.safe_load(f)
    if not validator.validate(config):
        raise ConfigValidationException(validator.errors)
    return validator.document


def validate_module_config(schema, config):
    validator = Validator(schema)
    if not validator.validate(config):
        raise ConfigValidationException(validator.errors)
    return validator.document



from __future__ import print_function
import logging

from abc import ABCMeta, abstractmethod

from .. import exceptions


try:
    # Python 2
    from urlparse import urlparse
except ImportError:
    # Python 3
    from urllib.parse import urlparse


LOG = logging.getLogger(__name__)


def install_missing_requirements(module):
    """
    Some of the modules require external packages to be installed. This gets
    the list from the `REQUIREMENTS` module attribute and attempts to
    install the requirements using pip.
    :param module: Backend module
    :type module: ModuleType
    :return: None
    :rtype: NoneType
    """
    reqs = getattr(module, "REQUIREMENTS", [])
    if not reqs:
        LOG.info("Module %r has no extra requirements to install." % module)
        return
    import pkg_resources

    pkgs_installed = pkg_resources.WorkingSet()
    pkgs_required = []
    for req in reqs:
        if req.startswith("git+"):
            url = urlparse(req)
            params = {
                x[0]: x[1] for x in map(lambda y: y.split("="), url.fragment.split("&"))
            }
            try:
                pkg = params["egg"]
            except KeyError:
                raise exceptions.CannotInstallModuleRequirements(
                    "Package %r in module %r must include '#egg=<pkgname>'"
                    % (req, module)
                )
        else:
            pkg = req
        if pkgs_installed.find(pkg_resources.Requirement.parse(pkg)) is None:
            pkgs_required.append(req)
    if pkgs_required:
        from subprocess import check_call, CalledProcessError

        try:
            check_call(["/usr/bin/env", "pip", "install"] + pkgs_required)
        except CalledProcessError as err:
            raise exceptions.CannotInstallModuleRequirements(
                "Unable to install packages for module %r (%s): %s"
                % (module, pkgs_required, err)
            )


class BackendBase(object):
    __metaclass__ = ABCMeta

    def __init__(self, name, config):
        self.name = name
        self.config = config

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def get_val(self, key):
        pass

    def get_all(self):
        raise exceptions.NoBackendSupport()

    def get_val_default(self, key, default):
        """
        Get a value from the backend, but fall back to the default value if
        the key doesn't exist.
        """
        try:
            return self.get_val(key)
        except exceptions.KeyNotFoundException:
            pass
        LOG.info(
            "Key %r not found in backend %r. Falling back to default value.",
            key,
            self.name,
        )
        return default



import logging
from os import environ as env

from . import BackendBase
from ..config import BASE_MODULE_SCHEMA
from ..exceptions import KeyNotFoundException


LOG = logging.getLogger(__name__)

CONFIG_SCHEMA = BASE_MODULE_SCHEMA.copy()
CONFIG_SCHEMA.update({"prefix": dict(type="string", default="")})


def get_vars(prefix):
    return {key[len(prefix) :]: value for key, value in env.items() if key.startswith(prefix)}


class Backend(BackendBase):
    def connect(self):
        return True

    def disconnect(self):
        return True

    def get_val(self, key):
        key = "%s%s" % (self.config.get("prefix", ""), key)
        LOG.debug("Getting value of key %r from environment", key)
        try:
            return env[key]
        except KeyError:
            raise KeyNotFoundException("No environment variable exists with key %r." % key)

    def get_all(self):
        prefix = self.config.get("prefix", "")
        LOG.debug("Geting all values with prefix %r from environment", prefix)
        return get_vars(prefix)
